#!/usr/bin/env python3.7

#  What does this script do?
# This script takes a pair of read (R1 and R2) in the format of dataframe row
# and call mutaitons based on the reads


import pandas as pd
import re
from Bio.Seq import Seq

from TileSeqMut import posterior
# import posterior

class MutParser(object):

    def __init__(self, row, full_seq, cds_seq, seq_lookup, tile_s, tile_e, post_prob_cutoff, logging, mut_rate):
        """
        row: input row includes both reads from sam file
        full_seq: full sequence (including cds and padding sequence)
        cds_seq: coding sequence (includes stop codon)
        seq: read from json file - seq, cds_start, cds_end
        """
        self._seq = Seq(full_seq["seq"].values.item())
        self._cds = Seq(cds_seq)
        self._start_pos = full_seq.cds_start.values.item() # 1 posisiton
        self._end_pos = full_seq.cds_end.values.item()

        self._tile_begins = tile_s
        self._tile_ends = tile_e
        # Note all the position in this df are 1-based
        self._seq_lookup = seq_lookup

        # get paired reads and their cigar scores
        self._reads = row

        # user defined post_prob_cutoff
        self._cutoff = post_prob_cutoff
        self._mutrate = mut_rate

        # logger
        self._logging = logging

    def _get_seq(self):
        """
        Parse the sequence and CIGAR string
        For R1 and R2, get reference sequence
        """
        # get information for R1
        # where the mapping starts
        self._r1_pos = int(self._reads["pos_start_r1"])
        # self._r1_cigar = list(cigar.Cigar(self._reads["cigar_r1"]).items())
        r1_cigar = [i for i in re.split("([\d]+\S)", self._reads["cigar_r1"]) if i != ""]
        self._r1_cigar = [(re.split(r"([\d]+)", i)[1], re.split(r"([\d]+)", i)[2]) for i in r1_cigar]
        self._r1_readlen = sum([int(i[0]) for i in self._r1_cigar])
        self._r1_ref = self._seq[int(self._r1_pos)-1:int(self._r1_pos)+len(self._reads["seq_r1"])]
        self._r1_qual = self._reads["qual_r1"]
        self._r1_read = self._reads["seq_r1"]
        self._r1_mdz = self._reads["mdz_r1"]

        # get the ref sequence for R2
        self._r2_pos = int(self._reads["pos_start_r2"])
        # self._r2_cigar = list(cigar.Cigar(self._reads["cigar_r2"]).items())
        r2_cigar = [i for i in re.split("([\d]+\S)", self._reads["cigar_r2"]) if i != ""]
        self._r2_cigar = [(re.split(r"([\d]+)", i)[1], re.split(r"([\d]+)", i)[2]) for i in r2_cigar]
        self._r2_readlen = sum([int(i[0]) for i in self._r2_cigar])
        self._r2_ref = self._seq[int(self._r2_pos)-1:int(self._r2_pos)+len(self._reads["seq_r2"])]
        self._r2_qual = self._reads["qual_r2"]
        self._r2_read = self._reads["seq_r2"]
        self._r2_mdz = self._reads["mdz_r2"]

        return self

    def _convert_mut(self, mut_list, read):
        """
        convert a list of mutations into the input format for posterior function
        return a df with mutations
        """
        if mut_list == []:
            return pd.DataFrame({}, columns=["m", "read", "pos", "ref", "alt", "qual"])
        final_mut = []
        for i in mut_list:
            mut = i.split("|")
            if "del" in mut:
                del_pos = int(mut[0])
                for base in mut[1]:
                    m = f"{del_pos}|{base}|del|{mut[-1]}"
                    del_pos+=1
                    final_mut.append(m)
            else:
                final_mut.append(i)
        final_df = pd.DataFrame({"m":final_mut})
        final_df["read"] = read
        final_df[["pos", "ref", "alt", "qual"]] = final_df["m"].str.split("|", expand=True)
        # validate final_df
        # for each read, there shouldn't be duplicated positions
        if final_df.pos.duplicated().any():
            self._logging.error(f"duplicated positions presented")
            self._logging.error(final_df)
            raise DuplicatedMutations()

        # for all the mutations
        # there shouldn't be any ref == alt
        same_refalt = final_df[final_df["ref"] == final_df["alt"]]
        if not same_refalt.empty:
            raise SameRefandAlt(same_refalt)

        return final_df

    def _main(self):
        """
        return a list of mutations from paired reads (R1 and R2)
        """
        # assign names to items in the dictionary
        self._get_seq()

        # parse mutations in R1
        r1_snp, r1_delins, map_pos_r1 = self._parse_cigar_mdz(self._r1_cigar, self._r1_mdz, self._r1_ref, self._r1_read,
                                                 self._r1_pos,self._r1_qual)
        # parse mutations in R2
        r2_snp, r2_delins, map_pos_r2 = self._parse_cigar_mdz(self._r2_cigar, self._r2_mdz, self._r2_ref, self._r2_read,
                                                  self._r2_pos,self._r2_qual)

        final_df_r1 = self._convert_mut(r1_snp, "r1")
        final_df_r2 = self._convert_mut(r2_snp, "r2")

        # merge two df
        snp_df = pd.merge(final_df_r1, final_df_r2, on=["pos"], how="outer", suffixes=('_r1', '_r2'))
        snp_df["pos"] = pd.to_numeric(snp_df["pos"])

        final_df_r1 = self._convert_mut(r1_delins, "r1")
        final_df_r2 = self._convert_mut(r2_delins, "r2")

        # merge two df
        delins_df = pd.merge(final_df_r1, final_df_r2, on=["pos"], how="outer", suffixes=('_r1', '_r2'))
        delins_df["pos"] = pd.to_numeric(delins_df["pos"])

        merged_df = [snp_df, delins_df]
        merged_df = pd.concat(merged_df)
        merged_df = merged_df.sort_values(by="pos")

        # group mutations based on positions
        # any mutations that are within 3bp are grouped together
        n = 3
        d = dict(tuple(merged_df.groupby(merged_df['pos'].diff().gt(n).cumsum())))
        # analyze the dictionary of clusters
        # and get posterior
        # print(d, self._mutrate, self._cutoff)
        pos_df, all_df, clustered_r1, clustered_r2 = posterior.cluster(d, self._r1_qual,self._r2_qual, map_pos_r1,
                                                                 map_pos_r2, self._mutrate,
                                           self._cutoff)
        final_mut = list(set(pos_df.m.tolist()))
        final_mut.sort()

        final_r1_cluster = list(set(clustered_r1.m.tolist()))
        final_r1_cluster.sort()
        final_r2_cluster = list(set(clustered_r2.m.tolist()))
        final_r2_cluster.sort()

        if final_mut != []:
            hgvs, outside_mut = self._get_hgvs(final_mut)
        else:
            hgvs, outside_mut = [], []

        if final_r1_cluster != []:
            hgvs_r1_clusters, outside_mut_r1 = self._get_hgvs(final_r1_cluster)
        else:
            hgvs_r1_clusters, outside_mut_r1 = [], []

        if final_r2_cluster != []:
            hgvs_r2_clusters, outside_mut_r2 = self._get_hgvs(final_r1_cluster)
        else:
            hgvs_r2_clusters, outside_mut_r2 = [], []

        if "c.463T>G" in hgvs:
            print(merged_df)
            # print(pos_df)
            print(all_df)

        return hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters

    def _parse_cigar_mdz(self, cigar, mdz_raw, ref, read, pos, qual):
        """
        use CIGAR string and mdz tag to call mutations from a read
        """
        snp_list = [] # output list with all the mutations
        delins_list = []
        # soft clip
        clip = 0
        # check starting point of the sequence
        # remove soft clipped bases and adjust for position
        if cigar[0][1] == "S": # soft clip occurs in the beginning of a read
            clip += int(cigar[0][0])

        # convert mdz string into a list
        mdz = re.findall('.*?[.ATCG]+', mdz_raw)
        # convert cigar list to a string
        cigar_joined = [str(i[0])+str(i[1]) for i in cigar]

        ins_pos_ref = 0 # on ref sequence where was the insertion
        total_len = 0 # total lenth indicates in CIGAR string
        ins_pos = [] # store [ins_pos, ins_len]

        mapped = 0
        deleted = 0

        # create a dictionary to store the mapped positions (ref_pos:read_pos)
        # this is later used to get the quality score for a base
        map_pos = {}
        ref_positions = []
        read_positions = []
        read_start = 0+clip
        ref_start = pos
        #if "I" in "".join(cigar_joined): # if insertion is in the read
            # get insertion position from cigar string
        for i in cigar: # go through each cigar string
            if i[1] != "I": # not insertion
                ins_pos_ref += int(i[0])
                if i[1] == "M": # track number of bases mapped
                    ref_positions += list(range(ref_start, ref_start+int(i[0])))
                    read_positions += list(range(read_start, read_start+int(i[0])))
                    ref_start += int(i[0])
                    read_start += int(i[0])
                    mapped += int(i[0])

                if i[1] == "D": # track number of bases deleted
                    deleted += int(i[0])
                    #read_start += int(i[0])
                    # skip deleted bases on ref read
                    ref_start += int(i[0])

            else:
                # based on number of bases mapped, get the inserted base from read
                ins_base = read[read_start:read_start+int(i[0])]
                qual_base = qual[read_start:read_start + int(i[0])]
                qual_base = ",".join(list(qual_base))
                # calculate inserted base posterior by using the quality string
                # qual_ins = qual[mapped:mapped+int(i[0])]
                # keep the insertion position and inserted lenth in a list
                ins_pos.append([ins_pos_ref, int(i[0])])
                # add the insertion to mut_list
                #mut_list.append([str(pos+mapped+deleted),ins_base,"ins"])
                delins_list.append(f"{pos + mapped + deleted}|{ins_base}|ins|{qual_base}")
                # skip inserted bases on read
                read_start += int(i[0])
        # zip two lists into a dictionary
        pos_map = dict(zip(ref_positions, read_positions))
        # parse snp and deletion from mdz string
        # given the inserted position on reference sequence
        r = re.compile("([0-9]+)([a-zA-Z\^]+)")
        # create iterator to go through all the insertions
        iter_ins = iter(ins_pos)
        ins = next(iter_ins, None)

        read_pos = 0
        inserted_pos = 0
        deleted_len = 0
        map_pos = 0 + clip

        for i in mdz:
            # for each item in the MD:Z string
            # split the item into number and letter
            m = r.match(i)
            match_len = int(m.group(1))
            base = m.group(2)

            map_pos += match_len# update how many bp are mapped

            while ins and map_pos >= ins[0]:
                inserted_pos += ins[1]
                ins = next(iter_ins, None)

            if "^" not in base:
                read_pos += match_len
                # this means a single nt change
                # mut_list.append([str(pos+read_pos-clip),base,read[read_pos+inserted_pos-deleted_len],
                # qual[read_pos+inserted_pos-deleted_len]])
                pos_on_read = read_pos+inserted_pos-deleted_len+clip
                snp_list.append(f"{pos+read_pos}|{base}|{read[pos_on_read]}|{qual[pos_on_read]}")

                # adjust read pos with 1bp change (move to the right for 1 pos)
                map_pos += len(base)
                read_pos += 1
            else: # deletion
                read_pos += match_len
                #mut_list.append([str(pos+read_pos-clip),base[1:],"del"])
                # quality of the base before deleted base
                qual_pre = qual[read_pos+inserted_pos-deleted_len+clip-1]
                # quality of the base after deleted base
                qual_post = qual[read_pos+inserted_pos-deleted_len+clip]
                delins_list.append(f"{pos + read_pos}|{base[1:]}|del|{qual_pre},{qual_post}")

                deleted_len += len(base[1:])
                read_pos += len(base[1:])
                map_pos += len(base[1:])

        # if "287|T|C|E" in snp_list:
        #     print(snp_list, delins_list)
        #     print(self._r1_pos,
        #     self._r1_cigar,
        #     self._r1_readlen,
        #     self._r1_ref,
        #     self._r1_qual,
        #     self._r1_read ,
        #     self._r1_mdz)
        #
        #     print(self._r2_pos,
        #           self._r2_cigar,
        #           self._r2_readlen,
        #           self._r2_ref,
        #           self._r2_qual,
        #           self._r2_read,
        #           self._r2_mdz)

        return snp_list, delins_list, pos_map

    def _get_hgvs(self, mut_list):
        """
        mut_list mutation list in the format
        translate a list of mut to aa changes
        return list of nt changes represent by hgvs strings
        """
        # if the snp changes are concecutive, it will be represent as delins
        concecutive_snp = []
        combined_snp = ""

        # track mutation positions that was not within the tiled region
        outside_mut = []

        # use to track delins
        # if del and ins happened within 2bp from each other then we consider it a delins
        # also record any snp that happened in the range
        delins = []
        mutations = []

        for mut in mut_list:

            mut_change = mut.split("|")
            tmp_pos = int(mut_change[0]) # template position

            # get cds position from lookup table
            try:
                cds_pos = self._seq_lookup[self._seq_lookup.temp_pos == tmp_pos].cds_pos.values.item()
            except:
                # this means that the mutation was on the template
                continue

            if cds_pos < self._tile_begins or cds_pos > self._tile_ends:
                #self._logging.warning(f"mutation at pos {cds_pos} which is not within the tile")
                outside_mut.append(cds_pos)
                continue

            if "N" in mut: # do not consider base N
                continue

            if ("del" in mut) or ("ins" in mut): # deletion or insertion
                mut_change[0] = cds_pos

                if delins == []: # nothing is on track
                    delins.append(mut_change)
                else: # compare bases
                    # delins = [[19,T,del]]
                    prev_pos = delins[-1][0] # previous deletion or insertion position
                    prev_bases = delins[-1][1]
                    prev_change = delins[-1][2]
                    if "del" in mut:
                        record_pos = mut_change[0]
                    elif "ins" in mut:
                        record_pos = mut_change[0] - 1

                    if (prev_pos-2 <= record_pos <= prev_pos+2) and (mut_change[2] != prev_change): # within 2bp of previous change
                        # not concecutive del or ins
                        # add this to delins
                        delins.append(mut_change)
                    else:
                        # output hgvs of mutations in delins
                        hgvs = delins_to_hgvs(self._cds, delins)
                        # update delins with current mutation
                        delins = [mut_change]
                        mutations.append(hgvs)

            else: # snp
                #mut_change = mut.split("|")
                #tmp_pos = int(mut_change[0])
                # get cds position from look up table
                #cds_pos = self._seq_lookup[self._seq_lookup.temp_pos == tmp_pos].cds_pos.item()
                cds_ref = self._cds[cds_pos-1]

                ## validate: if the reference base matches the ref in seq_lookup
                if cds_ref != mut_change[1]:
                    print(self._reads)
                    raise ValueError(f"Reference base - pos {tmp_pos}, base {mut_change[1]} does not match the reference provided by user - base {cds_ref}")
                # track concecutive changes
                if len(concecutive_snp) == 0:
                    concecutive_snp.append(cds_pos)
                    combined_snp+=mut_change[2]
                else:
                    if cds_pos == concecutive_snp[-1] +1:
                        # update position in concec_pos
                        combined_snp+=mut_change[2]
                        # update basesin combined_bases
                        concecutive_snp.append(cds_pos)
                    elif cds_pos == concecutive_snp[-1] +2:
                        # get middle ref
                        m_ref = self._cds[concecutive_snp[-1]+1-1]
                        concecutive_snp.append(cds_pos)
                        combined_snp += m_ref+mut_change[2]
                    elif cds_pos > concecutive_snp[-1] +2:
                        # convert things in concecutive_snp and combined_snp into hgvs
                        hgvs = snp_to_hgvs(concecutive_snp, combined_snp, self._cds)
                        # update combined_snp and hgvs with current mut
                        concecutive_snp = [cds_pos]
                        combined_snp = mut_change[2]
                        mutations.append(hgvs)

        if len(concecutive_snp) !=0:
            hgvs = snp_to_hgvs(concecutive_snp, combined_snp, self._cds)
            mutations.append(hgvs)
        if len(delins) != 0:
            hgvs = delins_to_hgvs(self._cds, delins)
            mutations.append(hgvs)
        if len(mutations) == 1:
            mutations = f"c.{mutations[0]}"
        elif len(mutations) == 0:
            return [], []
        else:
            joined = ";".join(mutations)
            mutations = f"c.[{joined}]"

        # output_df = pd.DataFrame(output_df, columns=["mut", "hgvs"])
        # print(output_df)

        return mutations, outside_mut


def snp_to_hgvs(concec_pos, combined_bases, cds):
    """
    helper function to obtain hgvs sstring given a list of positions and a string of bases combined.

    @param concec_pos
    @param combined_bases
    @param cds

    @return
    """
    if len(concec_pos) == 1:
        ref_base = cds[concec_pos[0]-1]
        hgvs = f"{concec_pos[0]}{ref_base}>{combined_bases}"
    else:
        hgvs = f"{concec_pos[0]}_{concec_pos[-1]}delins{combined_bases}"

    return hgvs


def delins_to_hgvs(cds_seq, delins):
    """
    convert a list of deletion and insertions (within 2 bp from eachother) to hgvs
    output hgvs should be ***_***delins***
    """
    # sort delins
    delins.sort()
    # convert delins to df
    delins_df = pd.DataFrame(delins, columns=["pos", "base", "type", "qual"])
    types = delins_df["type"].unique()

    if len(types) == 1:
        # only one type of mutations in the list
        # means the list has len = 1
        delins = delins[0]
        if types[0] == "ins": # insertion
            hgvs = f"{delins[0]-1}_{delins[0]}ins{delins[1]}"
        else: # deletion
            # if it is a one bp deletion
            if len(delins[1]) == 1:
                hgvs = f"{delins[0]}del"
            # it is more than one bp
            else:
                hgvs = f"{delins[0]}_{delins[0]+len(delins[1])-1}del"

    else: # means that the len is >1
        start_pos = delins[0][0]
        end_pos = delins[-1][0]
        prev_pos = 0
        modified = ""
        for index, row in delins_df.iterrows():
            if index == 0:
                # if the first change is deletion
                if row["type"] == "del":
                    modified += cds_seq[row["pos"]:row["pos"]+len(row["base"])]
                    prev_pos += row["pos"] + len(row["base"])
                else:
                    modified += row["base"]
                    prev_pos += row["pos"]-1
            else:
                if row["type"] == "del":
                    # first we need to get whatever is between this change and previous change
                    mid = cds_seq[prev_pos: row["pos"]-len(row["base"])]
                    modified += mid
                    prev_pos += len(mid)
                else:
                    mid = cds_seq[prev_pos: row["pos"]]
                    modified += mid+row["base"]
                    prev_pos += len(mid)

        hgvs = f"{start_pos}_{end_pos}delins{modified}"
        #print("del+ins, ",hgvs)
    return hgvs


class DuplicatedMutations(Exception):
    """Exception raised when there are two mutations found on the same read at the same positions.

        Attributes:
            message -- explanation of the error
        """

    def __init__(self, message="Same position mutated twice in one read!"):
        self.message = message
        super().__init__(self.message)


class SameRefandAlt(Exception):
    """Exception raised when reference base and alt base are the same.

        Attributes:
            message -- explanation of the error
        """

    def __init__(self, mut_df, message="Same reference base and alt base!"):
        self.mut = mut_df
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.mut} -> {self.message}'

if __name__ == "__main__":

    # test delins_to_hgvs
    cds_seq = "CATCTT"
    print(cds_seq[1:3])
    delins = [[2, "A", "del"], [4, "GG", "ins"], [5, "T", "del"]]
    #delins = [[2, "G", "ins"], [4, "C", "del"]]
    delins_to_hgvs(cds_seq, delins)



#!/usr/bin/env python3.7

# The functions in this script is used to manage submitting jobs to different clusters
# This script is called by main.py
import pandas as pd
import os
import re
import subprocess
import time

# other modules
from TileSeqMut import alignment

# import alignment

def alignment_sh_guru(fastq_map, ref_name, ref_seq, ref_path, sam_path, ds_sam_path, sh_output):
    """
    fastq_map: df contains paths to fastq files and downsamled fastq files
    ref_name: name for the reference sequence (same as project name)
    ref_seq: reference sequence to put in fasta file
    ref_path: build fasta in this path
    sam_path: path to sam files
    ds_sam_path: path to downsampled sam files
    sh_output: make sh files to submit to the cluster

    make reference for bowtie2
    make SGE sh files
    submit sh files to SGE
    """
    ## build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)

    for index, row in fastq_map.iterrows():
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, sample_name+"_aln.sh") # for each sample, the alignment is for both R1 and R2 (they are aligning separately)
        log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile)

        sub_cmd = f"qsub -cwd -N {'aln_'+sample_name} -e {log_file} {shfile}"
        os.system(sub_cmd)
        shfile_ds = os.path.join(sh_output, sample_name+"_aln_ds.sh")
        log_file_ds = alignment.align_main(ref, row["r1_ds"], row["r2_ds"], ds_sam_path, shfile_ds)
        sub_cmd = f"qsub -cwd -N {'aln_ds_'+sample_name} -e {log_file_ds} {shfile_ds}"
        os.system(sub_cmd)
        #break

def alignment_sh_bc2(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging, rc):
    """
    submit jobs to BC/BC2
    return a df with columns: [R1, R2, r1_sam, r2_sam]
    return a list of job id that we just submited
    """

    # build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)
    # store sam paths
    fastq_map = pd.concat([fastq_map, pd.DataFrame(columns=["r1_sam", "r2_sam"])])
    time = at
    all_job_id = []
    for index, row in fastq_map.iterrows(): # go through all the fastq pairs
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, f"Aln_{sample_name}.sh")
        r1_sam, r2_sam, log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile, rc=rc)
        row["r1_sam"] = r1_sam
        row["r2_sam"] = r2_sam
        # create log file for alignment
        sam_log_f = os.path.join(sam_path, f"{sample_name}.log")
        sub_cmd = ["submitjob2", "-w", str(time), "-c", "3", str(shfile), "2>", sam_log_f]
        jobs = subprocess.run(sub_cmd, stdout=subprocess.PIPE)

        job_id = jobs.stdout.decode("utf-8").strip().split(".")[0]
        all_job_id.append(job_id)
        logging.info(f"{sample_name}: job id - {job_id}")

    return fastq_map, all_job_id

def alignment_sh_dc(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging, rc):
    """
    submit jobs to BC/BC2
    return a df with columns: [R1, R2, r1_sam, r2_sam]
    return a list of job id that we just submited
    """

    # build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)
    # store sam paths
    fastq_map = pd.concat([fastq_map, pd.DataFrame(columns=["r1_sam", "r2_sam"])])
    time = at
    all_job_id = []
    for index, row in fastq_map.iterrows(): # go through all the fastq pairs
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, f"Aln_{sample_name}.sh")
        r1_sam, r2_sam, log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile, rc=rc)
        row["r1_sam"] = r1_sam
        row["r2_sam"] = r2_sam
        # create log file for alignment
        sam_log_f = os.path.join(sam_path, f"{sample_name}.log")
        sub_cmd = ["submitjob","-w", str(time),"-c", "3", str(shfile), "2>", sam_log_f]
        jobs = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = jobs.stdout.decode("utf-8").strip().split(".")[0]
        all_job_id.append(job_id)
        logging.info(f"Sample {sample_name}: job id - {job_id}")

    return fastq_map, all_job_id


def mut_count_sh_bc(sample_name, cmd, mt, sh_output_dir,logger, cores):
    """
    Submit mutation count jobs to BC

    """
    # go through files df and submit jobs for each pair of sam files
    # counting mutations in raw sam output files
    shfile = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.sh")
    with open(shfile, "w") as sh:
        sh.write(cmd+"\n")
        os.system(f"chmod 755 {shfile}")
    # submit this to the cluster
    sub_cmd = ["submitjob2","-w", str(mt), "-c", f"{cores}", "-m", "25", shfile]
    logger.debug(sub_cmd)
    job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
    job_id = job.stdout.decode("utf-8").strip().split(".")[0]
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id


def mut_count_sh_dc(sample_name, cmd, mt, sh_output_dir, logger, cores):
    """
    Submit mutation count jobs to DC
    """
    # go through files df and submit jobs for each pair of sam files
    # counting mutations in raw sam output files
    shfile = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.sh")
    with open(shfile, "w") as sh:
        sh.write(cmd+"\n")
        os.system(f"chmod 755 {shfile}")
    #sample_error_file = os.path.join(log_dir, f"sample_{sample_name}.log")
    # submit this to the cluster
    sub_cmd = ["submitjob", "-w", str(mt), "-c", f"{cores}", "-m", "25", shfile]
    logger.debug(sub_cmd)
    job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
    job_id = job.stdout.decode("utf-8").strip()
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id


def parse_jobs(job_list, env, logger):
    """
    return true if all the jobs in job list finished
    else wait for 10 mins and return how man jobs are running and queued
    job_list: list of job ids
    logger: logging object
    """
    qstat_cmd = ["qstat"] + job_list
    job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    qstat_out = job.stdout.decode("utf-8")
    qstat_err = job.stderr.decode("utf-8")

    f_id = []
    updated_list = []
    while True:
        running = []
        queued = []
        completed = []

        if qstat_err != "":
            # jobs might be finished and no longer in queue
            # extract job id from std err
            if env == "BC" or env == "BC2":
                err = qstat_err.split("\n")[:-1]
                id_regex = re.compile(r"(\d+).bc")
            elif env == "DC":
                err = qstat_err.split("\n")[:-1]
                id_regex = re.compile(r"(\d+).dc[0-9]+")

            f_id = [] # finished jobs
            for i in err:
                try:
                    match = id_regex.search(i)
                    job_id = match.group(1)
                    f_id.append(job_id)
                except:
                    print(i)
                    continue
            err_id = set(f_id)
            updated_list = [x for x in job_list if x not in err_id]

        if qstat_out != "":
            qstat_out = qstat_out.split("\n")[:-1]
            if env == "BC" or env == "BC2":
                id_regex = re.compile(r"(\d+).bc.+(R|Q|C|E)")
            elif env == "DC":
                id_regex = re.compile(r"(\d+).dc[0-9]+.+(R|Q|C|E)")

            for line in qstat_out:
                if ("---" in line) or ("Job ID" in line): continue
                match = id_regex.search(line)
                job_id = match.group(1)
                job_s = match.group(2)
                if job_s == "E" or job_s == "C":
                    completed.append(job_id)
                elif job_s == "R":
                    running.append(job_id)
                elif job_s == "Q":
                    queued.append(job_id)

        logger.info(f"{len(queued)} jobs queued")
        logger.info(f"{len(running)} jobs running")
        final_list = list(set(updated_list+running+queued))
        if final_list == []:
            return True
        else:
            # check in 10min
            time.sleep(600)
            job_list = final_list
            qstat_cmd = ["qstat"] + job_list
            job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            qstat_out = job.stdout.decode("utf-8")
            qstat_err = job.stderr.decode("utf-8")

if __name__ == "__main__":
    # test job list
    job_list = ["344775", "344777", "344771", "344780"]
    parse_jobs(job_list, "DC", "")



#!/usr/bin/env python3.7

# Author: Roujia Li
# email: Roujia.li@mail.utoronto.ca

import sys

sys.path.append('..')

import os
import pandas as pd

import multiprocessing as mp

def process_wrapper(combined_dict):
  """
  Call another module
  """
  return combined_dict

def main(f1, f2, cores):

  # open files and read lines
  file1 = open(f1, "r")
  file2 = open(f2, "r")

  # init pool objects
  pool = mp.Pool(cores)
  jobs = []
  combined = {}

  for line_f1, line_f2 in zip(file1, file2):
    # some preprocess of the lines
    line_f1 = line_f1.split()
    line_f2 = line_f2.split()

    if line_f1[0] != line_f2[0]:
      exit(1)

    if "#" in line_f1:
      continue

    # add some information from both lines into a dictionary
    combined["ID_f1"] = line_f1[0]
    combined["ID_f2"] = line_f2[0]

    combined["qual_f1"] = line_f1[3]
    combined["qual_f2"] = line_f2[3]

    # pass this dictionary to the process
    jobs.append(pool.apply_async(process_wrapper, (combined)))

  file1.close()
  file2.close()

  # get output from all the jobs and store the output
  id_list, name_list = [], []
  for job in jobs:
    ids, names = job.get()
    id_list.append(ids)
    name_list.append(names)

  # clean up
  pool.close()

if __name__ == '__main__':
  f1 = "/home/rothlab2/rli/tileseq_output/201011_LDLR_Surface_2020-10-15-17-54-25/sam_files/LDLR-Wildtype-T06-Rep2_90k_R1_001.sam"
  f2 = "/home/rothlab2/rli/tileseq_output/201011_LDLR_Surface_2020-10-15-17-54-25/sam_files/LDLR-Wildtype-T06-Rep2_90k_R2_001.sam"

  main(f1, f2, 8)


#!/usr/bin/env python3.7

## Read sam file (R1 and R2)
# count mutations in sam files
# output mutation counts

import pandas as pd
import re
import os
import math
import logging
import mmap
import multiprocessing as mp
import argparse
# import datetime
from collections import deque

# modules in package
from TileSeqMut import help_functions
from TileSeqMut import locate_mut

# import help_functions
# import locate_mut

class readSam(object):

    def __init__(self, sam_r1, sam_r2, param, arguments, output_dir, cores):
        """
        sam_R1: read one of the sample
        sam_R2: read two of the sample

        output_dir: main output directory

        log_level: settings for logging
        log_dir: directory to save all the logging for each sample
        """
        self._r1 = sam_r1
        self._r2 = sam_r2
        self._project, self._seq, self._cds_seq, self._tile_map, self._region_map, self._samples, self._var = help_functions.parse_json(param)

        self._qual = self._var["posteriorThreshold"]
        min_cover = self._var["minCover"]
        self._mutrate = self._var["mutRate"]
        self._output_counts_dir = output_dir
        self._cores = cores

        # sample information
        self._sample_id = os.path.basename(sam_r1).split("_")[0]
        self._sample_info = self._samples[self._samples["Sample ID"] == self._sample_id]

        # tile information
        self._sample_tile = self._sample_info["Tile ID"].values[0]
        self._tile_begins = (self._tile_map[self._tile_map["Tile Number"] == self._sample_tile]["Start AA"].values[0] *3)-2 # beginning position of this tile (cds position)
        self._tile_ends = self._tile_map[self._tile_map["Tile Number"] == self._sample_tile]["End AA"].values[0] *3  # ending position of this tile (cds position)
        self._tile_len = self._tile_ends - self._tile_begins
        self._cds_start = self._seq.cds_start
        self._min_map_len = math.ceil(self._tile_len * float(min_cover))

        self._sample_condition = self._sample_info["Condition"].values[0]
        self._sample_tp = self._sample_info["Time point"].values[0]
        self._sample_rep = self._sample_info["Replicate"].values[0]

        self._seq_lookup = help_functions.build_lookup(self._cds_start.values.item(), self._seq.cds_end.values.item(), self._cds_seq)

        # init a new object for logging, log to sample specific file.log
        log_f = os.path.join(output_dir, f"sample_{str(self._sample_id)}_mut_count.log")
        log_object = help_functions.logginginit(arguments.log_level, log_f)

        self._sample_counts_f = os.path.join(self._output_counts_dir,f"counts_sample_{self._sample_id}.csv")
        self._sample_counts_r1_f = os.path.join(self._output_counts_dir, f"counts_sample_{self._sample_id}_r1.csv")
        self._sample_counts_r2_f = os.path.join(self._output_counts_dir, f"counts_sample_{self._sample_id}_r2.csv")

        self._mut_log = log_object.getLogger("count.mut")
        self._locate_log = log_object.getLogger("locate.mut")

        self._mut_log.info(f"Counting mutations in sample-{self._sample_id}")
        self._mut_log.info(f"Sam file input R1:{sam_r1}")
        self._mut_log.info(f"Sam file input R2:{sam_r2}")

        output_csv = open(self._sample_counts_f, "w")
        # write log information to counts output
        output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                         f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                         f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                         f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
        output_csv.close()

        output_csv = open(self._sample_counts_r1_f, "w")
        # write log information to counts output
        output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                         f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                         f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                         f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
        output_csv.close()

        output_csv = open(self._sample_counts_r2_f, "w")
        # write log information to counts output
        output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                         f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                         f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                         f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
        output_csv.close()


    def _merged_main(self):
        """
        Read two sam files at the same time, store mutations that passed filter
        """
        read_pair = 0 # total pairs
        un_map = 0 # total number of unmapped reads
        read_nomut = 0 # read pairs that have no mutations

        hgvs_output = {} # save all the hgvs in this pair of sam files
        off_mut = {} # save all the positions that were mapped but outside of the tile
        row = {} # create a dictionary to save all the reads information to pass on to locate_mut.py

        final_pairs = 0
        off_read = 0

        chunkSize = 1500000 # number of characters in each chunk (you will need to adjust this)
        chunk1 = deque([""]) #buffered lines from 1st file
        chunk2 = deque([""]) #buffered lines from 2nd file
        with open(self._r1, "r") as r1_f, open(self._r2, "r") as r2_f:
            while chunk1 and chunk2:
                line_r1 = chunk1.popleft()
                if not chunk1:
                    line_r1,*more = (line_r1+r1_f.read(chunkSize)).split("\n")
                    chunk1.extend(more)
                line_r2 = chunk2.popleft()
                if not chunk2:
                    line_r2,*more = (line_r2+r2_f.read(chunkSize)).split("\n")
                    chunk2.extend(more)

                if line_r1.startswith("@") or line_r2.startswith("@"): # skip header lines
                    continue

                line_r1 = line_r1.split()
                line_r2 = line_r2.split()
                if len(line_r1) <= 1:
                    continue

                read_pair += 1 # count how many read pairs in this pair of sam files
                mapped_name_r1 = line_r1[2]
                mapped_name_r2 = line_r2[2]
                if mapped_name_r1 == "*" or mapped_name_r2 == "*": # if one of the read didn't map to ref
                    un_map +=1
                    continue

                # check if read ID mapped
                read_name_r1 = line_r1[0]
                read_name_r2 = line_r2[0]
                if read_name_r1 != read_name_r2:
                    print(read_name_r1)
                    print(read_name_r2)
                    self._mut_log.error("Read pair IDs did not map, please check fastq files")
                    exit(1)

                # get starting position for r1 and r2
                pos_start_r1 = line_r1[3]
                pos_start_r2 = line_r2[3]
                # record reads that are not mapped to this tile
                # this is defined as if the read covers at least min_map percent of the tile
                # the default min_map is 70%
                # we also assume that the read len is ALWAYS 150
                r1_end = int(pos_start_r1) + 150
                # r1_end must cover from start of the tile to 70% of the tile
                if (r1_end - int(self._cds_start)) < (int(self._tile_begins) + int(self._min_map_len)):
                    off_read += 1
                    continue
                if (int(pos_start_r2) - int(self._cds_start)) > (int(self._tile_ends) - int(self._min_map_len)):
                    off_read += 1
                    continue
                # get CIGAR string
                CIGAR_r1 = line_r1[5]
                seq_r1 = line_r1[9]
                quality_r1 = line_r1[10]

                CIGAR_r2 = line_r2[5]
                seq_r2 = line_r2[9]
                quality_r2 = line_r2[10]

                mdz_r1 = [i for i in line_r1 if "MD:Z:" in i]
                mdz_r2 = [i for i in line_r2 if "MD:Z:" in i]

                if len(mdz_r1) != 0:
                    mdz_r1 = mdz_r1[0].split(":")[-1]
                else:
                    mdz_r1 = ""

                if len(mdz_r2) != 0:
                    mdz_r2 = mdz_r2[0].split(":")[-1]
                else:
                    mdz_r2 = ""

                if ((not re.search('[a-zA-Z]', mdz_r1)) and ("I" not in CIGAR_r1)) and ((not re.search('[a-zA-Z]', mdz_r2)) and ("I" not in CIGAR_r2)):
                    # if MDZ string only contains numbers
                    # and no insertions shown in CIGAR string
                    # means there is no mutation in this read
                    # if both reads have no mutations in them, skip this pair
                    read_nomut +=1
                    # remove reads that have no mutations in MDZ
                    continue

                # make the reads in the format of a dictionary
                # columns=["mapped_name", "pos_start", "qual", "CIGAR", "mdz","seq"])

                #row["mapped_name_r1"] = mapped_name_r1
                row["pos_start_r1"] = pos_start_r1
                row["qual_r1"] = quality_r1
                row["cigar_r1"] = CIGAR_r1
                row["mdz_r1"] = mdz_r1
                row["seq_r1"] = seq_r1

                #row["mapped_name_r2"] = mapped_name_r2
                row["pos_start_r2"] = pos_start_r2
                row["qual_r2"] = quality_r2
                row["cigar_r2"] = CIGAR_r2
                row["mdz_r2"] = mdz_r2
                row["seq_r2"] = seq_r2

                # pass this dictionary to locate mut
                # mut = locate_mut_main()
                # add mutation to mut list
                mut_parser = locate_mut.MutParser(row, self._seq, self._cds_seq, self._seq_lookup, self._tile_begins, self._tile_ends, self._qual, self._locate_log, self._mutrate)
                hgvs, outside_mut, pos_df, all_posterior= mut_parser._main()
                if len(hgvs) !=0:
                    final_pairs +=1
                    if hgvs in hgvs_output:
                        hgvs_output[hgvs] += 1
                    else:
                        hgvs_output[hgvs] = 1
                    #hgvs_output.append(hgvs)
                if outside_mut != []:
                    outside_mut = list(set(outside_mut))
                    for i in outside_mut:
                        if not (i in off_mut):
                            off_mut[i] = 1

        # track mutations that are not within the same tile
        # track sequencing depth for each sample
        # write this information to a tmp file
        # merge the tmp file (in main.py)
        tmp_f = os.path.join(self._output_counts_dir,f"{self._sample_id}_tmp.csv")
        off_mut = list(set(off_mut))
        off_mut.sort()
        f = open(tmp_f, "w")
        f.write(f"sample,{self._sample_id},tile,{self._tile_begins}-{self._tile_ends},sequencing_depth,{read_pair},off_tile_reads,{off_read},off_tile_perc,{off_read/read_pair}\n")
        f.close()

        output_csv = open(self._sample_counts_f, "a")
        output_csv.write(f"#Raw read depth:{read_pair}\n")
        output_csv.write(f"#Number of read pairs without mutations:{read_nomut}\n#Number of read pairs did not map to gene:{un_map}\n#Mutation positions outside of the tile:{off_mut}\n#Number of reads outside of the tile:{off_read}\n")
        output_csv.write(f"#Final read-depth:{read_pair - un_map - off_read}\n")
        output_csv.write(f"#Total read pairs with mutations:{final_pairs}\n")
        output_csv.write(f"#Comment: Total read pairs with mutations = Read pairs with mutations that passed the posterior threshold\n#Comment: Final read-depth = raw read depth - reads didn't map to gene - reads mapped outside of the tile\n")

        output_csv.close()

        self._mut_log.info(f"Raw sequencing depth: {read_pair}")
        self._mut_log.info(f"Number of reads without mutations:{read_nomut}")
        self._mut_log.info(f"Final read-depth: {read_pair - un_map - off_read}")
        # convert list to df with one col
        hgvs_df = pd.DataFrame.from_dict(hgvs_output, orient="index")
        hgvs_df = hgvs_df.reset_index()
        hgvs_df.columns = ["HGVS", "count"]
        hgvs_df.to_csv(self._sample_counts_f, mode="a", index=False)

    def multi_core(self):
        """
        Read two sam files at the same time, store mutations that passed filter
        """
        read_pair = 0  # total pairs
        un_map = 0  # total number of unmapped reads
        read_nomut = 0  # read pairs that have no mutations

        hgvs_output = {}  # save all the hgvs in this pair of sam files
        r1_pop_hgvs = {}  # save all the hgvs that are 2/3nt changes based only on R1
        r2_pop_hgvs = {}  # save all the hgvs that are 2/3nt changes based only on R2
        off_mut = {}  # save all the positions that were mapped but outside of the tile
        row = {}  # create a dictionary to save all the reads information to pass on to locate_mut.py

        final_pairs = 0
        r1_popmut = 0
        r2_popmut = 0
        off_read = 0

        # chunkSize = 1500000  # number of characters in each chunk
        # chunk1 = deque([""])  # buffered lines from 1st file
        # chunk2 = deque([""])  # buffered lines from 2nd file
        r1_f = open(self._r1, "r")
        r2_f = open(self._r2, "r")
        # init objects
        pool = mp.Pool(self._cores)
        jobs = []
        self._mut_log.info("Start reading SAM files")
        for line_r1, line_r2 in zip(r1_f, r2_f):

            line_r1 = line_r1.split()
            line_r2 = line_r2.split()
            if len(line_r1) <= 1:
                continue

            read_pair += 1  # count how many read pairs in this pair of sam files
            mapped_name_r1 = line_r1[2]
            mapped_name_r2 = line_r2[2]
            if mapped_name_r1 == "*" or mapped_name_r2 == "*":  # if one of the read didn't map to ref
                un_map += 1
                continue

            # check if read ID mapped
            read_name_r1 = line_r1[0]
            read_name_r2 = line_r2[0]
            if read_name_r1 != read_name_r2:
                self._mut_log.error("Read pair IDs did not map, please check fastq files")
                exit(1)

            # get starting position for r1 and r2
            pos_start_r1 = line_r1[3]
            pos_start_r2 = line_r2[3]
            # record reads that are not mapped to this tile
            # this is defined as if the read covers at least min_map percent of the tile
            # the default min_map is 70%
            # we also assume that the read len is ALWAYS 150
            r1_end = int(pos_start_r1) + 150
            # r1_end must cover from start of the tile to 70% of the tile
            if (r1_end - int(self._cds_start)) < (int(self._tile_begins) + int(self._min_map_len)):
                off_read += 1
                continue
            if (int(pos_start_r2) - int(self._cds_start)) > (int(self._tile_ends) - int(self._min_map_len)):
                off_read += 1
                continue
            # get CIGAR string
            CIGAR_r1 = line_r1[5]
            seq_r1 = line_r1[9]
            quality_r1 = line_r1[10]

            CIGAR_r2 = line_r2[5]
            seq_r2 = line_r2[9]
            quality_r2 = line_r2[10]

            mdz_r1 = [i for i in line_r1 if "MD:Z:" in i]
            mdz_r2 = [i for i in line_r2 if "MD:Z:" in i]

            if len(mdz_r1) != 0:
                mdz_r1 = mdz_r1[0].split(":")[-1]
            else:
                mdz_r1 = ""

            if len(mdz_r2) != 0:
                mdz_r2 = mdz_r2[0].split(":")[-1]
            else:
                mdz_r2 = ""

            if ((not re.search('[a-zA-Z]', mdz_r1)) and ("I" not in CIGAR_r1)) and (
                    (not re.search('[a-zA-Z]', mdz_r2)) and ("I" not in CIGAR_r2)):
                # if MDZ string only contains numbers
                # and no insertions shown in CIGAR string
                # means there is no mutation in this read
                # if both reads have no mutations in them, skip this pair
                read_nomut += 1
                # remove reads that have no mutations in MDZ
                continue
            # make the reads in the format of a dictionary
            # columns=["mapped_name", "pos_start", "qual", "CIGAR", "mdz","seq"])
            # row["mapped_name_r1"] = mapped_name_r1
            row["pos_start_r1"] = pos_start_r1
            row["qual_r1"] = quality_r1
            row["cigar_r1"] = CIGAR_r1
            row["mdz_r1"] = mdz_r1
            row["seq_r1"] = seq_r1

            # row["mapped_name_r2"] = mapped_name_r2
            row["pos_start_r2"] = pos_start_r2
            row["qual_r2"] = quality_r2
            row["cigar_r2"] = CIGAR_r2
            row["mdz_r2"] = mdz_r2
            row["seq_r2"] = seq_r2
            # mut_parser = locate_mut.MutParser(row, self._seq, self._cds_seq, self._seq_lookup, self._tile_begins,
                                              # self._tile_ends, self._qual, self._locate_log, self._mutrate)
            jobs.append(pool.apply_async(process_wrapper, (row, self._seq, self._cds_seq, self._seq_lookup, self._tile_begins,
                                               self._tile_ends, self._qual, self._locate_log, self._mutrate)))
            row = {} # flush

        r1_f.close()
        r2_f.close()
        self._mut_log.info("File streamed to subprocesses, waiting for jobs to finish")
        # wait for all jobs to finish
        all_df = []

        for job in jobs:
            hgvs, outside_mut, all_posterior, hgvs_r1_clusters, hgvs_r2_clusters = job.get()
            if len(hgvs) != 0:
                final_pairs += 1
                if hgvs in hgvs_output:
                    hgvs_output[hgvs] += 1
                else:
                    hgvs_output[hgvs] = 1
                # hgvs_output.append(hgvs)

            if len(hgvs_r1_clusters) != 0:
                r1_popmut += 1
                if hgvs_r1_clusters in r1_pop_hgvs:
                    r1_pop_hgvs[hgvs_r1_clusters] += 1
                else:
                    r1_pop_hgvs[hgvs_r1_clusters] = 1

            if len(hgvs_r2_clusters) != 0:
                r2_popmut += 1
                if hgvs_r2_clusters in r2_pop_hgvs:
                    r2_pop_hgvs[hgvs_r2_clusters] += 1
                else:
                    r2_pop_hgvs[hgvs_r2_clusters] = 1

            if outside_mut != []:
                outside_mut = list(set(outside_mut))
                for i in outside_mut:
                    if not (i in off_mut):
                        off_mut[i] = 1
            all_df.append(all_posterior)

        # clean up
        pool.close()
        self._mut_log.info("Pool closed")
        # track mutations that are not within the same tile
        # track sequencing depth for each sample
        # write this information to a tmp file
        # merge the tmp file (in main.py)

        # tmp_f = os.path.join(self._output_counts_dir, f"{self._sample_id}_tmp.csv")
        # off_mut = list(set(off_mut))
        # off_mut.sort()
        # f = open(tmp_f, "w")
        # f.write(
        #     f"sample,{self._sample_id},tile,{self._tile_begins}-{self._tile_ends},sequencing_depth,{read_pair},off_tile_reads,{off_read},off_tile_perc,{off_read / read_pair}\n")
        # f.close()

        output_csv = open(self._sample_counts_f, "a")
        output_csv.write(f"#Raw read depth:{read_pair}\n")
        output_csv.write(
            f"#Number of read pairs without mutations:{read_nomut}\n#Number of read pairs did not map to gene:{un_map}\n#Mutation positions outside of the tile:{off_mut}\n#Number of reads outside of the tile:{off_read}\n")
        output_csv.write(f"#Final read-depth:{read_pair - un_map - off_read}\n")
        output_csv.write(f"#Total read pairs with mutations:{final_pairs}\n")
        output_csv.write(
            f"#Comment: Total read pairs with mutations = Read pairs with mutations that passed the posterior threshold\n#Comment: Final read-depth = raw read depth - reads didn't map to gene - reads mapped outside of the tile\n")

        output_csv.close()

        self._mut_log.info(f"Raw sequencing depth: {read_pair}")
        self._mut_log.info(f"Number of reads without mutations:{read_nomut}")
        self._mut_log.info(f"Final read-depth: {read_pair - un_map - off_read}")
        # convert list to df with one col
        hgvs_df = pd.DataFrame.from_dict(hgvs_output, orient="index")
        hgvs_df = hgvs_df.reset_index()
        hgvs_df.columns = ["HGVS", "count"]
        hgvs_df.to_csv(self._sample_counts_f, mode="a", index=False)

        r1_df = pd.DataFrame.from_dict(r1_pop_hgvs, orient="index")
        r1_df = r1_df.reset_index()
        if not r1_df.empty:
            r1_df.columns = ["HGVS", "count"]
        r1_df.to_csv(self._sample_counts_r1_f, mode="a", index=False)

        r2_df = pd.DataFrame.from_dict(r2_pop_hgvs, orient="index")
        r2_df = r2_df.reset_index()
        if not r2_df.empty:
            r2_df.columns = ["HGVS", "count"]
        r2_df.to_csv(self._sample_counts_r2_f, mode="a", index=False)
        #
        # self._mut_log.info(f"Reading posterior filtered dfs ...")
        #
        # # pos_df = pd.concat(pos_df)
        # posterior_f = os.path.join(self._output_counts_dir, f"{self._sample_id}_posprob.csv")
        # for df in pos_df:
        #     df.to_csv(posterior_f, mode='a', header=False, index=False)

        self._mut_log.info(f"Reading posterior unfiltered dfs ...")
        # all_df = pd.concat(all_df)
        all_f = os.path.join(self._output_counts_dir, f"{self._sample_id}_posprob_all.csv")
        for df in all_df:
            df.to_csv(all_f, mode='a', header=False, index=False)
        self._mut_log.info(f"Posterior df saved to files ... {all_f}")


def process_wrapper(row, seq, cds_seq, seq_lookup, tile_begins, tile_ends, qual, locate_log, mutrate):
    """

    """
    mut_parser = locate_mut.MutParser(row, seq, cds_seq, seq_lookup, tile_begins, tile_ends, qual, locate_log, mutrate)
    hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters = mut_parser._main()
    return hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters


# if __name__ == "__main__":
#
#     parser = argparse.ArgumentParser(description='TileSeq mutation counts (for sam files)')
#     parser.add_argument("-r1", "--read_1", help="sam file for R1", required=True)
#     parser.add_argument("-r2", "--read_2", help="sam file for R2", required=True)
#     parser.add_argument("-qual", "--quality", help="quality filter using posterior probability", default=0.99)
#     parser.add_argument("-o", "--output", help="Output folder that contains the directory: ./sam_files", required = True)
#     parser.add_argument("-log", "--log_level", help="Set log level: debug, info, warning, error, critical.", default = "debug")
#     parser.add_argument("-mutlog", "--log_dir", help="Directory to save all the log files for each sample")
#     parser.add_argument("-p", "--param", help="csv paramter file", required = True)
#     parser.add_argument("-min", "--min_cover", help="Minimum percentage required to be covered by reads", default = 0.4)
#
#     args = parser.parse_args()
#
#     sam_r1 = args.read_1
#     sam_r2 = args.read_2
#     qual_filter = float(args.quality)
#     log_dir = args.log_dir
#     min_map = float(args.min_cover)
#
#     out = args.output
#     param = args.param
#
#     # conver the csv file to json
#     # csv2json = os.path.abspath("src/csv2json.R")
#     param_json = param.replace(".csv", ".json")
#     #convert = f"Rscript {csv2json} {param} -o {param_json} -l stdout"
#     #os.system(convert)
#
#     # process the json file
#     project, seq, cds_seq, tile_map, region_map, samples = help_functions.parse_json(param_json)
#     # build lookup table
#     lookup_df = help_functions.build_lookup(seq.cds_start.values.item(), seq.cds_end.values.item(), cds_seq)
#
#     # initialize the object
#     MutCounts = readSam(sam_r1, sam_r2, seq, lookup_df, tile_map, region_map, samples, out, qual_filter, min_map, args.log_level, log_dir)
#
#     MutCounts._merged_main(seq, cds_seq)



#!/usr/bin/python3.7
# to run this script
# python random_sample.py -input /path/to/fastq.gz
# - all the files in the input directory have to be gzipped
# - the script output down sampled fastq files in the same directory as the original fastq files
# to move them to a new dir you can run `mv *_ds.fastq /new_dir/`
# output fastq files are NOT gzipped, you will need to zip them inorder to run song's script

import pandas as pd
import os
import random
import argparse


def random_lines(r1, r2, n):
    """
    get n random records from input file
    output to file *_ds.fastq
    """
    record_number = 0

    base_r1 = os.path.basename(r1)
    base_r2 = os.path.basename(r2)

    output_r1 = os.path.dirname(r1)+"/"+base_r1.replace(".fastq", "_ds.fastq")
    output_r2 = os.path.dirname(r2)+"/"+base_r2.replace(".fastq", "_ds.fastq")

    with open(r1, "r") as i:
        r1_lines = sum([1 for line in i])
    with open(r2, "r") as j:
        r2_lines = sum([1 for line in j])

    total_records_r1 = int(r1_lines/4)
    total_records_r2 = int(r2_lines/4)

    if n > total_records_r1 or n > total_records_r2:
        os.system("cp "+r1+ " "+output_r1)
        os.system("cp "+r2+ " "+output_r2)
        return None
    else:
        if total_records_r1 <= total_records_r2:
            records_to_keep = set(random.sample(range(total_records_r1 + 1), n))
        else:
            records_to_keep = set(random.sample(range(total_records_r2 + 1), n))

    with open(r1, "r") as i_1:
        with open(output_r1, "w") as o1:

            for line1 in i_1:
                line2 = i_1.readline()
                line3 = i_1.readline()
                line4 = i_1.readline()
                if record_number in records_to_keep:
                    o1.write(line1)
                    o1.write(line2)
                    o1.write(line3)
                    o1.write(line4)
                record_number += 1


    record_number = 0
    with open(r2, "r") as i_2:
        with open(output_r2, "w") as o2:
            for line1 in i_2:
                line2 = i_2.readline()
                line3 = i_2.readline()
                line4 = i_2.readline()
                if record_number in records_to_keep:
                    o2.write(line1)
                    o2.write(line2)
                    o2.write(line3)
                    o2.write(line4)
                record_number += 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("-input", help="input folder contains original fastq files")
    parser.add_argument("-n", help="number of lines to downsample to")

    args = parser.parse_args()

    n = args.n

    for f in os.listdir(args.input):
        if f.endswith(".fastq.gz") and ("_R1_" in f):

            r2 = f.replace("_R1_", "_R2_")

            r1 = os.path.join(args.input, f)
            r2 = os.path.join(args.input, r2)

            os.system("gunzip "+r1)
            os.system("gunzip "+r2)

            r1 = r1.replace(".gz", "")
            r2 = r2.replace(".gz", "")

            random_lines(r1, r2, n)



#!/usr/bin/env python3.7

# Author: Roujia Li
# email: Roujia.li@mail.utoronto.ca

import os
import pandas as pd
import json
import argparse
from TileSeqMut.main import log

#for two runs merge the mutation counts files into output folder


def merge_runs(input_dir1, input_dir2, map_df, output):
	"""
	:param input_dir1: Directory 1 contains mutation counts files
	:param input_dir2: Directory 2 contains mutation counts files
	:param map_df: A dataframe that maps the samples to be merged
	:param output: output directory to save merged mut_counts
	"""
	# make a new file

	new_samples = []
	for index, row in map_df.iterrows():
		merged_header = ""
		mut_count_f1 = os.path.join(input_dir1, f"counts_sample_{row['Sample ID_run1']}.csv")
		mut_count_f2 = os.path.join(input_dir2, f"counts_sample_{row['Sample ID_run2']}.csv")
		new_samples.append(f"{row['Sample ID_run1']}-{row['Sample ID_run2']}")
		# join headers
		with open(mut_count_f1) as file1, open(mut_count_f2) as file2:
			for line1, line2 in zip(file1, file2):
				if "#" not in line1: break
				if "Final read-depth:" in line1:
					n_depth_1 = int(line1.split(":")[1])
					n_depth_2 = int(line2.split(":")[1])
					total_depth = n_depth_1+n_depth_2
					merged_header+= f"#Final read-depth:{total_depth}\n"
					continue
				merged_header += line1.strip() + "; " + line2
		output_f = open(os.path.join(output, f"counts_sample_{row['Sample ID_run1']}-{row['Sample ID_run2']}.csv"), "w")
		output_f.write(merged_header)
		# skip header and read the mutations
		df1 = pd.read_csv(mut_count_f1, skiprows=18)
		df2 = pd.read_csv(mut_count_f2, skiprows=18)
		merge = [df1, df2]
		merge = pd.concat(merge)
		merge_counts = merge.groupby(by=["HGVS"], as_index=False)["count"].sum().sort_values("count", ascending=False)
		# save header and counts to file
		merge_counts.to_csv(output_f, mode="a", index=False)
		output_f.close()

	return new_samples


def parse_params(param1, param2, output, main_log):
	"""
	Given two parameter sheets, find samples in common and output
	df for later merging
	:param param1: parameter sheet for the first run
	:param param2: parameter sheet for the second run
	:param main_log: logger
	:return map_df: df with two cols [samples_1, samples_2]
	"""
	update_json = {}
	with open(param1, "r") as p1:
		with open(param2, "r") as p2:
			param_one = json.load(p1)
			param_two = json.load(p2)
			samples1 = param_one["samples"]
			samples2 = param_two["samples"]
			samples1_df = pd.DataFrame(samples1)
			samples2_df = pd.DataFrame(samples2)
	# from these two files find samples in the same tile, with the same condition names
	# same replicates and tp
	# log how many samples will be merged
	merge = pd.merge(samples1_df, samples2_df, how="inner", on=["Tile ID", "Condition", "Time point", "Replicate"],
					 suffixes=["_run1", "_run2"])
	unique_Tid = merge["Tile ID"].drop_duplicates().tolist()
	unique_Conditions = merge["Condition"].drop_duplicates().tolist()
	merge["new_samples"] = merge["Sample ID_run1"] + "-" + merge["Sample ID_run2"]
	# save new sample names to output
	merge[["new_samples", "Tile ID", "Condition", "Time point", "Replicate"]].to_csv(os.path.join(output, "new_sample_names.csv"), index=False)
	main_log.info(f"{merge.shape[0]} samples will be merged")
	main_log.info(f"Tiles: {unique_Tid}")
	main_log.info(f"Condition: {unique_Conditions}")

	return merge


def merge_main(args):

	# check args and make corresponding output dir
	# log input/output information
	if not os.path.isdir(args.output):
		os.mkdir(args.output)
	logging = log(args.output, args.log_level)
	main_log = logging.getLogger("mergeRuns.log")
	main_log.info(f"Input dir 1: {args.dir1} with Parameter Sheet: {args.paramOne}")
	main_log.info(f"Input dir 2: {args.dir2} with Parameter Sheet: {args.paramTwo}")
	main_log.info(f"Merges mutation count file will be saved to {args.output}")

	map_df = parse_params(args.paramOne, args.paramTwo, args.output, main_log)
	new_samples = merge_runs(args.dir1, args.dir2, map_df, args.output)

	main_log.info(f"Merged files are saved to {args.output}")

	# make a json file with merged sample names (copy the other information from old json files)
	new_paramsheet = ""

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description='TileSeq mutation counts')
	# user input arguments
	parser.add_argument("-p1", "--paramOne", help="Path to parameter sheet (JSON) for the first run", type=str, required=True)
	parser.add_argument("-p2", "--paramTwo", help="Path to parameter sheet (JSON) for the second run", type=str, required=True)
	parser.add_argument("-d1", "--dir1", help="Path to *_mut_count folder for the first run", type=str, required=True)
	parser.add_argument("-d2", "--dir2", help="Path to *_mut_count folder for the second run", type=str, required=True)
	parser.add_argument("-o", "--output", help="Output folder path (this folder will be made if not exists when "
											   "running the script",
						type=str, required=True)

	parser.add_argument("-log", "--log_level", help="set log level: debug, \
	        info, warning, error, critical. (default = debug)", type=str, default="debug")

	args = parser.parse_args()

	merge_main(args)



#!/usr/bin/env python3.7
import glob
import math
import sys
import os
import pandas as pd
import numpy as np
import argparse
from ast import literal_eval
from matplotlib_venn import venn2
from fpdf import FPDF
import progressbar

from matplotlib import rc

import seaborn as sns
import warnings
from scipy import stats
import matplotlib.pyplot as plt
sys.path.append('..')
from TileSeqMut import help_functions
# import help_functions
warnings.simplefilter(action='ignore')
# Author: Roujia Li
# email: Roujia.li@mail.utoronto.ca

# QC script for posteriors


class PosteriorQC(object):

    def __init__(self, mut_count_dir, param_json, log_level="INFO"):
        """
        @param mut_count_dir: Directory ends with _mut_count
        @param param_json: Json parameter sheet
        """
        # validate input parameters
        if not os.path.isdir(mut_count_dir):
            raise NotADirectoryError(f"Dir does not exist: {mut_count_dir}")

        if not os.path.isfile(param_json):
            raise FileNotFoundError(f"File does not exist: {param_json}")

        self._project, self._seq, self._cds_seq, self._tile_map, \
            self._region_map, self._samples, self._var = help_functions.parse_json(param_json)

        self._mut_dir = mut_count_dir
        self._output = os.path.join(mut_count_dir, "posterior_qc")

        if not os.path.isdir(self._output):
            os.mkdir(self._output)

        log_object = help_functions.logginginit(log_level, "posteriorQC.log")
        self._logger = log_object.getLogger("posteriorQC")

        self._logger.info(f"Reading files from {os.path.abspath(self._mut_dir)}")
        self._logger.info(f"Reading parameters from {os.path.abspath(param_json)}")
        self._logger.info(f"Saving output plots to {os.path.abspath(self._output)}")
        self._i = 0

    def prob_analysis(self, row, pdf, bar):
        """
        @param row: row of dataframe contains path to prob files and sample info
        """
        # load prob_df
        fig_title = f"{row['Condition']}_tile{row['Tile ID']}_rep{row['Replicate']}_t{row['Time point']}"
        # prob_df = pd.read_csv(row.prob, names=["mut", "prob", "read"], header=None)
        # print(prob_df)
        # self._plot_prob(prob_df)
        # self._plot_marginal(prob_df)
        sns.set(font_scale=1.7)
        sns.set_style("whitegrid")
        fig, axes = plt.subplots(3, 2, figsize=(16, 18))
        axes[-1, -1].axis('off')
        all_prob_df = pd.read_csv(row.prob_all, names=["mut", "prob", "read", "pass"], header=None)
        pop_r1_df = pd.read_csv(row.popmut_r1, skiprows=9)
        pop_r2_df = pd.read_csv(row.popmut_r2, skiprows=9)
        self._plot_prob(all_prob_df, fig_title, fig, axes)
        self._plot_r1r2_popmut(pop_r1_df, pop_r2_df, fig_title, fig, axes)
        # plt.tight_layout(pad=0.6)
        fig.suptitle(fig_title, fontweight='bold')
        fig.tight_layout(pad=1)
        plt.savefig(os.path.join(self._output, f"{fig_title}.png"), format="png")
        plt.close()
        pdf.add_page()
        pdf.image(os.path.join(self._output, f"{fig_title}.png"), 2, 25, 200, 235)
        bar.update(self._i + 1)
        self._i +=1
        # self._logger.info(f"Processed sample {fig_title}")

        # plot_name = f'Tile{row["Tile ID"]}_{row["Condition"]}_t{row["Time point"]}_rep{row["Replicate"]}.png'
        # self._plot_counts(prob_df, plot_name)
        # plot_name = f'Tile{row["Tile ID"]}_{row["Condition"]}_t{row["Time point"]}_rep{row["Replicate"]}_all.png'
        # self._plot_counts(prob_df, plot_name)

    def _plot_prob(self, df, fig_title, fig, axes):
        """
        Plot probability of read 1 and read 2
        """

        # mutations only on read 1
        read1_only = df.loc[df["read"] == "r1"]

        # mutations only on read 2
        read2_only = df.loc[df["read"] == "r2"]

        read1_only.prob = read1_only.prob.astype(float)
        read2_only.prob = read2_only.prob.astype(float)
        read1_only_marginal = self._calculate_frequency(read1_only)
        read2_only_marginal = self._calculate_frequency(read2_only)

        map_r1 = {"1": "R1_Passed", "-1": "R1_Discarded"}
        map_r2 = {"1": "R2_Passed", "-1": "R2_Discarded"}

        read1_only = read1_only.replace({"pass": map_r1})
        read2_only = read2_only.replace({"pass": map_r2})

        top90_cut = np.log(0.9)
        top50_cut = np.log(0.5)
        # plot counts vs avg prob

        sns.scatterplot(read1_only_marginal["marginal_freq"], np.log(read1_only_marginal["prob"]),
                        label="Read 1", ax=axes[0][1], alpha=0.7)
        sns.scatterplot(read2_only_marginal["marginal_freq"], np.log(read2_only_marginal["prob"]),
                        label="Read 2", ax=axes[0][1], alpha=0.7)
        max_x = max(read1_only_marginal["marginal_freq"].max(), read2_only_marginal["marginal_freq"].max())
        # max_x = math.floor(max_x*10)/10

        axes[0][1].axhline(top90_cut, ls='--', alpha=.7, color="#FA5D5D")
        axes[0][1].axhline(top50_cut, ls='--', alpha=.7, color="#FA5D5D")

        axes[0][1].text((max_x/3)*2, top50_cut-0.38, "posterior = 0.5", fontsize=10, color="grey")
        axes[0][1].text((max_x/3)*2, top90_cut-0.38, "posterior = 0.9", fontsize=10, color="grey")

        axes[0][1].set_ylabel("Average posterior prob. (log)")
        axes[0][1].set_xlabel("Marginal frequency")
        axes[0][1].set_title("SNVs found only on R1 or R2")

        # axes[0][0].get_legend().remove()
        # plt.show()

        # plt.close()
        # plot marginal
        # sns.scatterplot(read1_only_marginal.index, read1_only_marginal.marginal_freq, s=20, ax=axes[0][0],
        #                 label="Read 1", linewidth=0.000001, alpha = 0.6)
        # sns.scatterplot(read2_only_marginal.index, read2_only_marginal.marginal_freq, s=20, ax=axes[0][0],
        #                 label="Read 2", linewidth=0.000001, alpha = 0.6)
        #
        # axes[0][0].set_title("Marginal frequencies of mutations found only on R1 or R2")
        # axes[0][0].set_xlabel("Unique variants")
        # axes[0][0].set_ylabel("Marginal Frequency")

        # plt.title(fig_title)
        # plt.savefig(os.path.join(self._output, f"{fig_title}_marginal_r1r2.pdf"))
        # plt.close()
        # mutations found on both reads
        both_read = df.loc[(df["read"] != "r1") & (df["read"] != "r2")]
        both_read['read'] = [literal_eval(x) for x in both_read['read']]
        both_read['prob'] = [literal_eval(x) for x in both_read['prob']]
        both_read[["prob_r1", "prob_r2"]] = pd.DataFrame(both_read['prob'].tolist(), index=both_read.index)
        both_read[["label_r1", "label_r2"]] = pd.DataFrame(both_read['read'].tolist(), index=both_read.index)
        both_read["pass"] = "Discarded"
        both_read["label"] = "both"
        both_read.loc[(both_read.prob_r1 > 0.5) & (both_read.prob_r2 > 0.5), "pass"] = "Passed"
        read1_only["log_prob"] = np.log(read1_only.prob)
        read2_only["log_prob"] = np.log(read2_only.prob)

        read1_only = read1_only.sort_values(by="prob", ascending=False).reset_index()
        read2_only = read2_only.sort_values(by="prob", ascending=False).reset_index()
        read1_only["label"] = "r1_only"
        read2_only["label"] = "r2_only"

        read1_only["perc"] = read1_only.index / (read1_only.shape[0] - 1)
        read2_only["perc"] = read2_only.index / (read2_only.shape[0] - 1)

        # unique variants r1 only
        unique_r1_only = read1_only.drop_duplicates(subset=["mut"])
        # unique variants r2 only
        unique_r2_only = read2_only.drop_duplicates(subset=["mut"])
        # unique variants both
        unique_both = both_read.drop_duplicates(subset=["mut"])
        #
        # print(read1_only.shape, read2_only.shape, both_read.shape)
        #
        # print(unique_r1_only.shape, unique_r2_only.shape, unique_both.shape)


        total_both = unique_both[["label", "pass"]].value_counts().to_frame()
        total_r1 = unique_r1_only[["label", "pass"]].value_counts().to_frame()
        total_r2 = unique_r2_only[["label", "pass"]].value_counts().to_frame()
        bar_plot_df = pd.concat([total_both, total_r2, total_r1]).reset_index()
        bar_plot_df.columns = ["label", "pass", "count"]

        map = {"R1_Passed": "Passed", "R1_Discarded": "Discarded", "R2_Passed": "Passed", "R2_Discarded": "Discarded"}
        color_dict = {"Passed": "#45736A", "Discarded": "#D94E4E"}

        bar_plot_df = bar_plot_df.replace({"pass": map})
        g = sns.barplot(x="label", y="count", hue="pass", data=bar_plot_df, ax=axes[0][0], palette=color_dict)
        for p in g.patches:
            height = p.get_height()
            if math.isnan(height):
                height=0
            g.text(p.get_x() + p.get_width() / 2.,
                    height + 3,
                    int(height),
                    ha="center")
        handles, labels = axes[0][0].get_legend_handles_labels()
        axes[0][0].legend(handles=handles, labels=labels)
        axes[0][0].set_xlabel("")
        axes[0][0].set_title("Number of unique mutations (SNVs) found")

        # join r1 and r2
        join_R1R2 = [read1_only, read2_only]
        join_R1R2_df = pd.concat(join_R1R2)

        # Create an array with the colors you want to use
        colors = ["#6BA3F2", "#3866A6"]
        # Set your custom color palette
        customPalette = sns.set_palette(sns.color_palette(colors))
        #
        colors = ["#F2C46B", "#A67B28"]
        customPalette2 = sns.set_palette(sns.color_palette(colors))
        color_dict = {"R1_Passed": "#6BA3F2", "R1_Discarded": "#3866A6", "R2_Passed": "#F2C46B",
                      "R2_Discarded":"#A67B28"}
        sns.scatterplot(x="perc", y="log_prob", hue="pass", palette=color_dict, data=join_R1R2_df,
                        ax=axes[1][0], s=20, style="pass", linewidth=0.000001, alpha = 0.7)
        handles, labels = axes[1][0].get_legend_handles_labels()
        axes[1][0].legend(handles=handles, labels=labels)
        # sns.scatterplot(x="perc", y="log_prob", hue="pass", palette=color_dict, data=read2_only,
        #                 ax=axes[1][0], s=20, style="pass", linewidth=0.000001, alpha = 0.7)
        # sns.scatterplot(x="perc", y="prob", hue="pass", data=read2_only , label=f"Read 2: n = {read2_only.shape[0]}",
        #                 ax=ax1, s=20)
        axes[1][0].axhline(top90_cut, ls='--', alpha=.7, color="#FA5D5D")
        axes[1][0].axhline(top50_cut, ls='--', alpha=.7, color="#FA5D5D")

        axes[1][0].text(0.6666, top50_cut-0.38, "posterior = 0.5", fontsize=10, color="grey")
        axes[1][0].text(0.6666, top90_cut-0.38, "posterior = 0.9", fontsize=10, color="grey")

        # plot prob correlation of r1 and r2
        sns.scatterplot(np.log(both_read.prob_r1), np.log(both_read.prob_r2), s=10, ax=axes[1][1], color="black")
        pcc = stats.pearsonr(both_read.prob_r1, both_read.prob_r2)[0]
        axes[1][0].set_title("Posterior prob of SNVs only found on R1 or R2")
        axes[1][0].set_ylabel("Posterior prob (log)")
        axes[1][0].set_xlabel("Proportion of SNVs")

        axes[1][1].set_title("Corr. of posterior prob of SNVs found on both reads")
        axes[1][1].text(-3, -1, f"PCC: {float('{:.4f}'.format(pcc))}")
        axes[1][1].set_ylabel("Posterior prob (log) (Read 1)")
        axes[1][1].set_xlabel("Posterior prob (log) (Read 2)")

        # exit()

        # # make plot for mutation on r1 only and r2 only vs all
        # n_r1 = read1_only.mut.value_counts().to_frame().reset_index()
        # n_r2 = read2_only.mut.value_counts().to_frame().reset_index()
        # print(n_r1)
        # print(n_r2)
        #
        # sns.scatterplot(n_r1.index, n_r1["mut"], label="Read 1")
        # sns.scatterplot(n_r2.index, n_r2["mut"], label="Read 2")
        # # bar_df = pd.DataFrame({"type": ["Mutations on read1", "Mutations on read2", "Mutations on both reads"],
        # #                        "count": [n_r1, n_r2, both]})
        # # sns.barplot(x="type", y="count", data=bar_df, palette=["#F07D78", "#B7F86E", "#316FA3"])
        # # plt.title("Number of mutations (before filtering) on each read")
        # plt.show()
        # plt.close()
        # exit()

    def _plot_r1r2_popmut(self, df_r1, df_r2, fig_title, fig, axes):
        """
        For R1 and R2, plot the distribution of popcode mutations based on the margional frequency
        """
        # print(df_r1)
        # print(df_r2)
        if df_r1.empty:
            df_r1 = pd.DataFrame({}, columns=["HGVS", "count"])
        if df_r2.empty:
            df_r2 = pd.DataFrame({}, columns=["HGVS", "count"])
        total_r1 = df_r1["count"].sum()
        df_r1["marginal freq"] = df_r1["count"] / total_r1
        total_r2 = df_r2["count"].sum()
        df_r2["marginal freq"] = df_r2["count"] / total_r2

        # sort by freq
        # df_r1 = df_r1.sort_values(by="marginal freq", ascending=False)
        # df_r2 = df_r2.sort_values(by="marginal freq", ascending=False)
        #
        # sns.scatterplot(df_r1["HGVS"], df_r1["marginal freq"], ax = axes[2][0])
        # sns.scatterplot(df_r2["HGVS"], df_r2["marginal freq"], ax = axes[2][1])
        # axes[2][0].set_xticklabels("")
        # axes[2][1].set_xticklabels("")
        # axes[2][0].set_facecolor('white')
        # axes[2][1].set_facecolor('white')

        # get union set of hgvs
        all_hgvs = set(df_r1["HGVS"].tolist()+df_r2["HGVS"].tolist())
        intersect  = [value for value in df_r1["HGVS"].tolist() if value in df_r2["HGVS"].tolist()]

        unique_r1 = df_r1.loc[~df_r1.HGVS.isin(df_r2)].dropna()
        unique_r2 = df_r2.loc[~df_r2.HGVS.isin(df_r1)].dropna()

        # sort by freq
        unique_r1 = unique_r1.sort_values(by="marginal freq", ascending=False)
        unique_r2 = unique_r2.sort_values(by="marginal freq", ascending=False)

        if len(all_hgvs) != 0:
            j_dist = len(intersect) / len(all_hgvs)
            venn2([set(unique_r1.HGVS.tolist()), set(unique_r2.HGVS.tolist())], set_labels=('Read 1', 'Read 2'),
                  ax=axes[2][0])
            axes[2][0].set_title("Unique 2/3nt changes in R1 or R2 (passed threshold cutoff)")
            axes[2][0].text(.5, .6, f"Jaccard Index: {round(j_dist, 2)}")

        else: # no 2/3nt changes found on either reads
            axes[2][0].axis("off")

        # sns.scatterplot(unique_r1["HGVS"], unique_r1["marginal freq"], ax = axes[2][0])
        # sns.scatterplot(unique_r2["HGVS"], unique_r2["marginal freq"], ax = axes[2][1])
        # axes[2][0].set_xticklabels("")
        # axes[2][1].set_xticklabels("")
        #
        # axes[2][0].set_facecolor('white')
        # axes[2][1].set_facecolor('white')
        #
        # print(fig_title)
        # print(len(all_hgvs))
        # print(len(intersect))
        # print(len(intersect)/len(all_hgvs))


    def _plot_marginal(self, df):
        """

        """
        # mutations only on read 1
        read1_only = df.loc[df["read"] == "r1"]
        # mutations only on read 2
        read2_only = df.loc[df["read"] == "r2"]

        # mutations found on both reads
        both_read = df.loc[(df["read"] != "r1") & (df["read"] != "r2")]
        both_read['read'] = [literal_eval(x) for x in both_read['read']]
        both_read['prob'] = [literal_eval(x) for x in both_read['prob']]
        both_read[["prob_r1", "prob_r2"]] = pd.DataFrame(both_read['prob'].tolist(), index=both_read.index)
        both_read[["label_r1", "label_r2"]] = pd.DataFrame(both_read['read'].tolist(), index=both_read.index)

        r1_marginal = self._calculate_frequency(read1_only)
        r1_marginal["read"] = "r1"
        r2_marginal = self._calculate_frequency(read2_only)
        r2_marginal["read"] = "r2"
        both_read_marginal = self._calculate_frequency(both_read)
        both_read_marginal["read"] = "both"

        marginal_concat = pd.concat([r1_marginal, r2_marginal, both_read_marginal])
        sns.violinplot(x = "read", y="marginal_freq", data=marginal_concat)
        plt.show()
        plt.close()

    def _calculate_frequency(self, df):
        """
        Calculate marginal frequencies for all the variants in the table
        """
        # count each unique variant in the table
        mut_counts = df["mut"].value_counts()
        mut_counts = mut_counts.to_frame().reset_index()
        mut_counts.columns = ["mut", "count"]
        total_mut = mut_counts["count"].sum()
        mut_counts["marginal_freq"] = mut_counts["count"] / total_mut
        avg_prob = df.groupby(['mut'])['prob'].mean().to_frame().reset_index()
        mut_counts = pd.merge(mut_counts, avg_prob, how="left", on="mut")
        return mut_counts

    def _plot_counts(self, df, f_name):
        """
        make stacked barplots of number of mutations on each read
        """
        # mutations only on read 1
        read1_only = df.loc[df["read"] == "r1"]
        # mutations only on read 2
        read2_only = df.loc[df["read"] == "r2"]

        # mutations found on both reads
        both_read = df.loc[(df["read"] != "r1") & (df["read"] != "r2")]
        both_read['read'] = [literal_eval(x) for x in both_read['read']]
        both_read['prob'] = [literal_eval(x) for x in both_read['prob']]
        both_read[["prob_r1", "prob_r2"]] = pd.DataFrame(both_read['prob'].tolist(), index=both_read.index)
        both_read[["label_r1", "label_r2"]] = pd.DataFrame(both_read['read'].tolist(), index=both_read.index)

        # make plot for mutation on r1 only and r2 only vs all
        n_r1 = read1_only.shape[0]
        n_r2 = read2_only.shape[0]
        both = both_read.shape[0]

        bar_df = pd.DataFrame({"type": ["Mutations on read1", "Mutations on read2", "Mutations on both reads"],
                               "count": [n_r1, n_r2, both]})
        sns.barplot(x="type", y="count", data=bar_df, palette=["#F07D78", "#B7F86E", "#316FA3"])
        plt.title("Number of mutations (before filtering) on each read")
        plt.savefig(os.path.join(self._output, f_name))
        plt.close()

    def runQC(self):
        """
        Run posterior QC, generate posterior QC plots in a subdir
        """
        # go through sample ID list and find corresponding files
        f_table = []
        # join samples table into one col with _ delim
        for sample in self._samples["Sample ID"].tolist():
            # filtered prob df
            # filtered_prob_f = os.path.join(self._mut_dir, f"{sample}_posprob.csv")
            # if not os.path.isfile(filtered_prob_f):
            #     raise FileNotFoundError(f"{sample}_posprob.csv")
            all_prob_f = os.path.join(self._mut_dir, f"{sample}_posprob_all.csv")
            if not os.path.isfile(all_prob_f):
                # raise FileNotFoundError(f"{sample}_posprob_all.csv")
                print(f"{sample}_posprob_all.csv")
                continue

            pop_r1_f = os.path.join(self._mut_dir, f"counts_sample_{sample}_r1.csv")
            pop_r2_f = os.path.join(self._mut_dir, f"counts_sample_{sample}_r2.csv")
            f_table.append([sample, all_prob_f, pop_r1_f, pop_r2_f])

        f_df = pd.DataFrame(f_table, columns=["Sample ID", "prob_all", "popmut_r1", "popmut_r2"])
        # merge this to sample df
        merge_file_samples = pd.merge(self._samples, f_df, on="Sample ID", how="left")

        # go through each condition to make QC plots
        conditions = list(set(merge_file_samples["Condition"]))
        total_sample = merge_file_samples.shape[0]
        self._logger.info(f"Total sampels = {total_sample}")
        bar = progressbar.ProgressBar(maxval=total_sample, widgets=[progressbar.Bar('=', '[', ']'), ' ',
                                                                    progressbar.Percentage()])
        bar.start()
        for c in conditions:
            # get number of subplots we have to make for this condition
            # fig, axes = plt.subplots(4, 2, figsize=(13, 30))
            c_df = merge_file_samples.loc[merge_file_samples["Condition"] == c]
            # sort df by tiles so the output plots are in the right order
            c_df = c_df.sort_values(by="Tile ID")
            # print(c_df["Sample ID"])
            # exit()
            pdf = FPDF('P', 'mm', 'A4')
            c_df.apply(lambda x: self.prob_analysis(x, pdf, bar), axis=1)
            #
            # # merge all the files for this condition
            # c_list = glob.glob(f"{self._output}/{c}*")
            # c_list.sort()
            # for image in c_list:
            #     pdf.add_page()
            #     pdf.image(image, 2, 25, 200, 245)

            pdf.output(f"{self._output}/{c}_posteriorQC.pdf", "F")
        bar.finish()
        os.system(f"rm {self._output}/*.png")
        self._logger.info("Posterior QC completed.")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='TileSeq mutation posterior QC')
    # user input arguments
    parser.add_argument("-i", "--input", help="Path to folder that contains mutation counts (*_mut_count)", type=str,
                        required=True)
    parser.add_argument("-p", "--param", help="json paramter file", type=str, required=True)

    args = parser.parse_args()

    p_qc = PosteriorQC(args.input, args.param)
    p_qc.runQC()



#!/usr/bin/env python3.7

# This is used to calculate the posterior probability for variants
# the script is modified from varCallerSnippet.r

import math
import pandas as pd
import numpy as np
from fractions import Fraction

def cluster(mut_cluster: dict, r1_qual: str, r2_qual: str, r1_mappos: dict,
                        r2_mappos: dict, mut_rate: float, cut_off: float):
    """
    Parse cluster of mutations
    mut_cluster is a  list of clusters (clustered mutations found in one read pair)
    Mutations are clustered together if they are within
    @param mut_cluster dictionary where the values are dfs with mutations from both r1 and r2
    @param mut_rate mut rate defined by user
    @param cut_off posterior cutoff defined by user
    """
    pos_prob = {"m": [], "prob": [], "read": []}
    all_prob = {"m": [], "prob": [], "read": [], "pass":[]}

    # this is used to record 2/3nt changes on r1 or r2
    clustered_r1_mut = [pd.DataFrame({"m": [], "prob": [], "read": []})]
    clustered_r2_mut = [pd.DataFrame({"m": [], "prob": [], "read": []})]
    for c in mut_cluster.keys():
        mutcall = mut_cluster[c]
        # each mutcall is a df with columns:
        # m_r1,read,pos,ref_r1,alt_r1,qual_r1,m_r2,ref_r2,alt_r2,qual_r2
        c_size = mutcall.shape[0]  # size of the cluster
        # iterate through the clusters
        pos = {}
        r = ""

        # use this df to save the temporary mutations in this cluster
        tmp_cluster_mut = {"m": [], "prob": [], "read":[]}
        for index, row in mutcall.iterrows():

            if (pd.isnull(row["m_r1"]) or row["alt_r1"] == "N") and not pd.isnull(row["ref_r2"]):

                if r1_mappos.get(int(row["pos"])) is None:
                    continue

                # get R1 wt qual with map pos and qual str
                r1_qual_base = r1_qual[r1_mappos[int(row["pos"])]]
                # calculate posterior prob
                pos = bayesian_variant_call([row["ref_r2"], row["alt_r2"]], [r1_qual_base, row["qual_r2"]],
                                            row["ref_r2"], mut_rate, c_size)

                # add this information to all prob df
                all_prob["m"].append(row["m_r2"])
                all_prob["prob"].append(pos[row["alt_r2"]])
                all_prob["read"].append("r2")

                # if pos[row["ref_r2"]] > pos[row["alt_r2"]]:
                #     # this means that wt has higher pos-prob, then we dump this mutation
                #     all_prob["pass"].append(-1)

                # if we only consider the probability of mutation that we got from r2
                # check if this is greater than the probability cutoff
                if pos[row["alt_r2"]] > cut_off and c_size > 1:
                    # add this information to tmp prob df
                    tmp_cluster_mut["m"].append(row["m_r2"])
                    tmp_cluster_mut["prob"].append(pos[row["alt_r2"]])
                    tmp_cluster_mut["read"].append("r2")

                if pos[row["alt_r2"]] > cut_off and pos[row["ref_r2"]] <= pos[row["alt_r2"]]:
                    all_prob["pass"].append(1)

                    pos_prob["m"].append(row["m_r2"])
                    pos_prob["prob"].append(pos[row["alt_r2"]])
                    pos_prob["read"].append("r2")
                else:
                    all_prob["pass"].append(-1)

                r = "r2"

            elif (pd.isnull(row["m_r2"]) or row["alt_r2"] == "N") and not pd.isnull(row["ref_r1"]):

                if r2_mappos.get(int(row["pos"])) is None:
                    continue
                # get R1 wt qual with map pos and qual str
                r2_qual_base = r2_qual[r2_mappos[int(row["pos"])]]

                pos = bayesian_variant_call([row["alt_r1"], row["ref_r1"]], [row["qual_r1"], r2_qual_base],
                                            row["ref_r1"],
                                            mut_rate,
                                            c_size)
                all_prob["m"].append(row["m_r1"])
                all_prob["prob"].append(pos[row["alt_r1"]])
                all_prob["read"].append("r1")

                # if pos[row["ref_r1"]] > pos[row["alt_r1"]]:
                #     all_prob["pass"].append(-1)

                # if we only consider the probability of mutation that we got from r2
                # check if this is greater than the probability cutoff
                if pos[row["alt_r1"]] > cut_off and c_size > 1:
                    # add this information to tmp prob df
                    tmp_cluster_mut["m"].append(row["m_r1"])
                    tmp_cluster_mut["prob"].append(pos[row["alt_r1"]])
                    tmp_cluster_mut["read"].append("r1")

                if pos[row["alt_r1"]] > cut_off and pos[row["ref_r1"]] <= pos[row["alt_r1"]]:
                    all_prob["pass"].append(1)

                    pos_prob["m"].append(row["m_r1"])
                    pos_prob["prob"].append(pos[row["alt_r1"]])
                    pos_prob["read"].append("r1")
                else:
                    all_prob["pass"].append(-1)
                r = "r1"

            elif (not pd.isnull(row["m_r2"])) and (not pd.isnull(row["ref_r1"])):

                basecall = [row["alt_r1"], row["alt_r2"]]
                qual = [row["qual_r1"], row["qual_r2"]]
                pos = bayesian_variant_call(basecall, qual, row["ref_r1"], mut_rate, c_size)
                all_prob["m"].append(row["m_r1"])
                all_prob["prob"].append((pos[row["alt_r1"]], pos[row["alt_r2"]]))
                all_prob["read"].append(("r1", "r2"))
                all_prob["pass"].append("-")

                # if we only consider the probability of mutation that we got from r1
                # check if this is greater than the probability cutoff
                if pos[row["alt_r1"]] > cut_off and c_size > 1:
                    # add this information to tmp prob df
                    tmp_cluster_mut["m"].append(row["m_r1"])
                    tmp_cluster_mut["prob"].append(pos[row["alt_r1"]])
                    tmp_cluster_mut["read"].append("r1")
                # if we only consider the probability of mutation that we got from r2
                # check if this is greater than the probability cutoff
                if pos[row["alt_r2"]] > cut_off and c_size > 1:
                    # add this information to tmp prob df
                    tmp_cluster_mut["m"].append(row["m_r2"])
                    tmp_cluster_mut["prob"].append(pos[row["alt_r2"]])
                    tmp_cluster_mut["read"].append("r2")

                # if (pos[row["ref_r1"]] > pos[row["alt_r1"]]) and (pos[row["ref_r2"]] > pos[row["alt_r2"]]): continue

            if r == "":
                if pos[row["alt_r1"]] > pos[row["alt_r2"]] and pos[row["alt_r1"]] > cut_off:
                    pos_prob["m"].append(row["m_r1"])
                    pos_prob["prob"].append(pos[row["alt_r1"]])
                    pos_prob["read"].append("r1")

                elif pos[row["alt_r2"]] > pos[row["alt_r1"]] and pos[row["alt_r2"]] > cut_off:
                    pos_prob["m"].append(row["m_r2"])
                    pos_prob["prob"].append(pos[row["alt_r2"]])
                    pos_prob["read"].append("r2")

                elif pos[row["alt_r1"]] == pos[row["alt_r2"]] and pos[row["alt_r1"]] > cut_off:
                    pos_prob["m"].append(row["m_r1"])
                    pos_prob["prob"].append((pos[row["alt_r1"]], pos[row["alt_r2"]]))
                    pos_prob["read"].append(("r1", "r2"))

            # check tmp prob df
            # if number of mutations passed on r1 or r2 greater than 1
            # output mutations, label them r1 and r2, also assign them to the same cluster
            tmp_cluster = pd.DataFrame(tmp_cluster_mut)
            r1_mut_cluster = tmp_cluster[tmp_cluster["read"] == "r1"]
            r2_mut_cluster = tmp_cluster[tmp_cluster["read"] == "r2"]
            if r1_mut_cluster.shape[0] > 1:
                clustered_r1_mut.append(r1_mut_cluster)
            if r2_mut_cluster.shape[0] > 1:
                clustered_r2_mut.append(r2_mut_cluster)

    # print(pos_prob)
    pos_df = pd.DataFrame(pos_prob)
    all_df = pd.DataFrame(all_prob)
    clustered_r1_mut = pd.concat(clustered_r1_mut)
    clustered_r2_mut = pd.concat(clustered_r2_mut)
    return pos_df, all_df, clustered_r1_mut, clustered_r2_mut


def bayesian_variant_call(basecall, qual, wt, mut_rate, clusterSize=1):
    """
    @param basecall: list of base calls (i.e R1 -> A R2 -> C :  ["A", "C"])
    @param phred: phred score for the base calls (in letters) ["!", "J"]
    @param wt: wild type base
    @param mut_rate: mutation rate
    return: dictinary with basecall as keys and post prob as values
    """
    # all possible hypo bases
    nt = list(set([wt]+basecall))
    # nt = ["A", "G", "C", "T"]
    # convert phred to int scores
    # convert phred to int scores
    # in the case of deletions or insertions, there are multiple phred scores for each mut call
    phred = []
    for i in qual:
        if len(i) == 1:
            phred.append(10 ** (-(ord(i) - 33) / 10))
        else:
            all_phred = [10 ** (-(ord(j) - 33) / 10) for j in i.split(",")]
            phred.append(np.prod(all_phred))

    # phred = [10**(-(ord(i) - 33) / 10) for i in phred]
    post_p = []
    for base in nt: # go through each nt
        log_odd = 0
        if base == wt:
            log_odd += math.log((1-mut_rate) ** clusterSize) - math.log(1-((1-mut_rate)**clusterSize)/3)
        elif base == "ins":
            log_odd += (math.log(mut_rate) - math.log(4)) - math.log(1-(mut_rate/4))
        else:
            log_odd += math.log(1-(1-mut_rate) ** clusterSize) - math.log(3) - math.log(1-(1-(1-mut_rate) ** clusterSize)/3)

        for j in range(len(basecall)):
            if basecall[j] == base:
                log_odd += (math.log(1-phred[j]) - math.log(phred[j]) + math.log(3))
            else:
                log_odd += (math.log(phred[j]) - math.log(3) - math.log((1/3) -(phred[j]/9)))
        # print(log_odd)
        # print(basecall)
        logit_value = math.exp(log_odd) / (1+math.exp(log_odd))
        post_p.append(logit_value)

    prob = dict(zip(nt, post_p))
    output = dict(zip(basecall, [prob.get(base) for base in basecall]))

    return output


if __name__ == "__main__":
    basecall = ["T", "A"]
    phred = ["I", "J"]
    wt = "C"
    mut_rate= 0.0025
    prob = bayesian_variant_call(basecall, phred, wt, mut_rate)
    print(prob)



#!/usr/bin/env python3.7

# THis script is used to align fatq files to reference sequence
# Input: list of fastq files
# Output: samfiles

import os

def make_ref(name, ref_seq, ref_path):
    """
    given the reference sequence and ref_path
    make fasta file and build from fasta
    """
    ref_fasta = os.path.join(ref_path, name+".fasta")
    with open(ref_fasta, "w") as fasta:
        fasta.write(">"+name+"\n")
        fasta.write(ref_seq+"\n")
    build_cmd =f"bowtie2-build --quiet -f {ref_fasta} {os.path.join(ref_path, name)}"
    os.system(build_cmd)

    return os.path.join(ref_path, name)

def align_main(ref, r1, r2, sam_path, shfile, rc=False):
    """
    ref: reference fatsa file
    r1: input fastq file - R1
    r2: input fastq file - R2
    sam_path: path to sam files
    bowtie2: path to bowtie2
    shfile: write the command to this file
    return shfile
    """
    log_file = os.path.join(sam_path, os.path.basename(shfile).replace(".sh", ".log"))

    r1_sam_file = os.path.join(sam_path, os.path.basename(r1).replace(".fastq.gz", ".sam"))
    r2_sam_file = os.path.join(sam_path, os.path.basename(r2).replace(".fastq.gz", ".sam"))

    if not rc: # do not need to check for reverse complement
        r1_cmd = f"bowtie2 -p 3 --no-head --norc --no-sq --rdg 12,1 --rfg 12,1 --local -x {ref} -U {r1} -S {r1_sam_file}"
        r2_cmd = f"bowtie2 -p 3 --no-head --nofw --no-sq --rdg 12,1 --rfg 12,1 --local -x {ref} -U {r2} -S {r2_sam_file}"
    else:
        r1_cmd = f"bowtie2 -p 3 --no-head --no-sq --rdg 12,1 --rfg 12,1 --local -x {ref} -U {r1} -S {r1_sam_file}"
        r2_cmd = f"bowtie2 -p 3 --no-head --no-sq --rdg 12,1 --rfg 12,1 --local -x {ref} -U {r2} -S {r2_sam_file}"

    with open(shfile, "w") as f:
        f.write(r1_cmd + "\n")
        f.write(r2_cmd + "\n")
    os.system(f"chmod 755 {shfile}")
    return r1_sam_file, r2_sam_file, log_file



#!/usr/bin/env python3.7

# Author: Roujia Li
# email: Roujia.li@mail.utoronto.ca

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

with open("README.md", "r") as fh:
    long_description = fh.read()

# python setup.py bdist_wheel --universal
# testpypi: twine upload --repository testpypi dist/*
# pip install --index-url https://test.pypi.org/simple/ TileSeqMut==

# pypi: twine upload dist/*
# pip install TileSeqMut


setup(
    name="TileSeqMut",
    version="0.3.901", # CHANGE RELEASE NOTE IN MAIN
    author="ROUJIA LI",
    author_email="roujia.li@mail.utoronto.ca",
    description="Analysis scriptsTileSeqMut for TileSeq sequencing data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RyogaLi/tilseq_mutcount",
	scripts=['bin/tileseq_mut', 'bin/mergeRuns', 'bin/posterior_QC', 'bin/random_ds'],
	# For a list of valid classifiers, see https://pypi.org/classifiers/
	classifiers=[  # Optional
		# How mature is this project? Common values are
		#   3 - Alpha
		#   4 - Beta
		#   5 - Production/Stable
		'Development Status :: 3 - Alpha',

		# Indicate who your project is intended for
		'Intended Audience :: Developers',
		'Topic :: Software Development :: Build Tools',

		# Pick your license as you wish
		'License :: OSI Approved :: MIT License',

		# Specify the Python versions you support here. In particular, ensure
		# that you indicate you support Python 3. These classifiers are *not*
		# checked by 'pip install'. See instead 'python_requires' below.
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
	],
	packages=find_packages(),  # Optional
	install_requires=['pandas', 'seaborn', 'numpy', 'biopython', 'fpdf', 'progressbar'],
    python_requires='>=3.7',

	# # To provide executable scripts, use entry points in preference to the
	# # "scripts" keyword. Entry points provide cross-platform support and allow
	# # `pip` to create the appropriate form of executable for the target
	# # platform.
	# #
	# # For example, the following would provide a command called `sample` which
	# # executes the function `main` from this package when invoked:
	# entry_points={  # Optional
	# 	'console_scripts': [
	# 		'sample=sample:main',
	# 	],
	# },
)